@@ -201,6 +201,8 @@ void tsm_screen_sb_down(struct tsm_screen *con, unsigned int num);
 void tsm_screen_sb_page_up(struct tsm_screen *con, unsigned int num);
 void tsm_screen_sb_page_down(struct tsm_screen *con, unsigned int num);
 void tsm_screen_sb_reset(struct tsm_screen *con);
+unsigned int tsm_screen_sb_get_line_count(struct tsm_screen *con);
+unsigned int tsm_screen_sb_get_line_pos(struct tsm_screen *con);
 
 void tsm_screen_set_def_attr(struct tsm_screen *con,
 			     const struct tsm_screen_attr *attr);
@@ -284,6 +286,26 @@ tsm_age_t tsm_screen_draw(struct tsm_screen *con, tsm_screen_draw_cb draw_cb,
 
 struct tsm_vte;
 
+/* terminal flags */
+#define TSM_VTE_FLAG_CURSOR_KEY_MODE			0x00000001 /* DEC cursor key mode */
+#define TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE		0x00000002 /* DEC keypad application mode; TODO: toggle on numlock? */
+#define TSM_VTE_FLAG_LINE_FEED_NEW_LINE_MODE		0x00000004 /* DEC line-feed/new-line mode */
+#define TSM_VTE_FLAG_8BIT_MODE				0x00000008 /* Disable UTF-8 mode and enable 8bit compatible mode */
+#define TSM_VTE_FLAG_7BIT_MODE				0x00000010 /* Disable 8bit mode and use 7bit compatible mode */
+#define TSM_VTE_FLAG_USE_C1				0x00000020 /* Explicitly use 8bit C1 codes; TODO: implement */
+#define TSM_VTE_FLAG_KEYBOARD_ACTION_MODE		0x00000040 /* Disable keyboard; TODO: implement? */
+#define TSM_VTE_FLAG_INSERT_REPLACE_MODE		0x00000080 /* Enable insert mode */
+#define TSM_VTE_FLAG_SEND_RECEIVE_MODE			0x00000100 /* Disable local echo */
+#define TSM_VTE_FLAG_TEXT_CURSOR_MODE			0x00000200 /* Show cursor */
+#define TSM_VTE_FLAG_INVERSE_SCREEN_MODE		0x00000400 /* Inverse colors */
+#define TSM_VTE_FLAG_ORIGIN_MODE			0x00000800 /* Relative origin for cursor */
+#define TSM_VTE_FLAG_AUTO_WRAP_MODE			0x00001000 /* Auto line wrap mode */
+#define TSM_VTE_FLAG_AUTO_REPEAT_MODE			0x00002000 /* Auto repeat key press; TODO: implement */
+#define TSM_VTE_FLAG_NATIONAL_CHARSET_MODE		0x00004000 /* Send keys from nation charsets; TODO: implement */
+#define TSM_VTE_FLAG_BACKGROUND_COLOR_ERASE_MODE	0x00008000 /* Set background color on erase (bce) */
+#define TSM_VTE_FLAG_PREPEND_ESCAPE			0x00010000 /* Prepend escape character to next output */
+#define TSM_VTE_FLAG_TITE_INHIBIT_MODE			0x00020000 /* Prevent switching to alternate screen buffer */
+
 /* keep in sync with shl_xkb_mods */
 enum tsm_vte_modifier {
 	TSM_SHIFT_MASK		= (1 << 0),
@@ -320,6 +342,62 @@ enum tsm_vte_color {
 	TSM_COLOR_NUM
 };
 
+/**
+ * Mouse Tracking
+ *
+ * Reference:
+ *
+ * https://invisible-island.net/xterm/ctlseqs/ctlseqs.html#h2-Mouse-Tracking
+ *
+ * The application running in the terminal can request a mouse tracking mode
+ * and it can configure the event type (send position on click only or on click
+ * and on mouse movement). The terminal will then send the position
+ * of the mouse cursor as requested by the application.
+ *
+ * Since libtsm doesn't know anything about the UI or the mouse this can only
+ * work if the terminal emulator built on top of libtsm cooperates.
+ *
+ * To implement mouse tracking a terminal emulator must first set a mouse
+ * callback with tsm_vte_set_mouse_cb. This callback will be called whenever the
+ * mouse mode changes in a way that is relevant to the terminal. It tells the
+ * terminal if it needs to pass mouse events on click, on move, or not at all.
+ *
+ * To pass the mouse events the terminal needs to call tsm_vte_handle_mouse with
+ * all parameters filled appropriately. The function will then take care of
+ * sending the mouse events to the application in the correct encoding and
+ * it will also discard events that are duplicate or unnecessary.
+ */
+
+/* control sequence codes sent be the application */
+#define TSM_VTE_MOUSE_MODE_X10      9 /* legacy mode (only cell mode, only on mouse click and x and y can be 223 max) */
+#define TSM_VTE_MOUSE_EVENT_BTN  1002 /* sends position on mouse click only */
+#define TSM_VTE_MOUSE_EVENT_ANY  1003 /* sends position on mouse click and mouse move */
+#define TSM_VTE_MOUSE_MODE_SGR   1006 /* modern mode that allows unlimited x and y coordinates */
+#define TSM_VTE_MOUSE_MODE_PIXEL 1016 /* sends pixel coordinates instead of cell coordinates */
+
+enum tsm_mouse_track_mode {
+	TSM_MOUSE_TRACK_DISABLE = 0, /* don't track mouse events */
+	TSM_MOUSE_TRACK_BTN = TSM_VTE_MOUSE_EVENT_BTN, /* call tsm_vte_handle_mouse only for mouse clicks */
+	TSM_MOUSE_TRACK_ANY = TSM_VTE_MOUSE_EVENT_ANY  /* call tsm_vte_handle_mouse for mouse clicks and mouse movement */
+};
+
+/* mouse buttons to be passed to tsm_vte_handle_mouse */
+#define TSM_MOUSE_BUTTON_LEFT       0
+#define TSM_MOUSE_BUTTON_MIDDLE     1
+#define TSM_MOUSE_BUTTON_RIGHT      2
+#define TSM_MOUSE_BUTTON_WHEEL_UP   4
+#define TSM_MOUSE_BUTTON_WHEEL_DOWN 5
+
+/* modifier keys to be passed to tsm_vte_handle_mouse (can be combined with OR) */
+#define TSM_MOUSE_MODIFIER_SHIFT  4
+#define TSM_MOUSE_MODIFIER_META   8
+#define TSM_MOUSE_MODIFIER_CTRL  16
+
+/* events to be passed to tsm_vte_handle_mouse */
+#define TSM_MOUSE_EVENT_PRESSED  1
+#define TSM_MOUSE_EVENT_RELEASED 2
+#define TSM_MOUSE_EVENT_MOVED    4
+
 typedef void (*tsm_vte_write_cb) (struct tsm_vte *vte,
 				  const char *u8,
 				  size_t len,
@@ -330,6 +408,11 @@ typedef void (*tsm_vte_osc_cb) (struct tsm_vte *vte,
 				  size_t len,
 				  void *data);
 
+typedef void (*tsm_vte_mouse_cb) (struct tsm_vte *vte,
+				  enum tsm_mouse_track_mode track_mode,
+				  bool track_pixels,
+				  void *data);
+
 int tsm_vte_new(struct tsm_vte **out, struct tsm_screen *con,
 		tsm_vte_write_cb write_cb, void *data,
 		tsm_log_t log, void *log_data);
@@ -337,6 +420,7 @@ void tsm_vte_ref(struct tsm_vte *vte);
 void tsm_vte_unref(struct tsm_vte *vte);
 
 void tsm_vte_set_osc_cb(struct tsm_vte *vte, tsm_vte_osc_cb osc_cb, void *osc_data);
+void tsm_vte_set_mouse_cb(struct tsm_vte *vte, tsm_vte_mouse_cb mouse_cb, void *mouse_data);
 
 /**
  * @brief Set color palette to one of the predefined palette on the vte object.
@@ -405,13 +489,33 @@ int tsm_vte_set_palette(struct tsm_vte *vte, const char *palette_name);
 int tsm_vte_set_custom_palette(struct tsm_vte *vte, uint8_t (*palette)[3]);
 
 void tsm_vte_get_def_attr(struct tsm_vte *vte, struct tsm_screen_attr *out);
+unsigned int tsm_vte_get_flags(struct tsm_vte *vte);
+
+unsigned int tsm_vte_get_mouse_mode(struct tsm_vte *vte);
+unsigned int tsm_vte_get_mouse_event(struct tsm_vte *vte);
 
 void tsm_vte_reset(struct tsm_vte *vte);
 void tsm_vte_hard_reset(struct tsm_vte *vte);
 void tsm_vte_input(struct tsm_vte *vte, const char *u8, size_t len);
+
+/**
+ * @brief Set backspace key to send either backspace or delete.
+ *
+ * Some terminals send ASCII backspace (010, 8, 0x08), some send ASCII delete
+ * (0177, 127, 0x7f).
+ *
+ * The default for vte is to send ASCII backspace.
+ *
+ * @param vte The vte object to set on
+ * @param enable Send ASCII delete if \c true, send ASCII backspace if \c false.
+ */
+void tsm_vte_set_backspace_sends_delete(struct tsm_vte *vte, bool enable);
 bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 			     uint32_t ascii, unsigned int mods,
 			     uint32_t unicode);
+bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
+        unsigned int cell_y, unsigned int pixel_x, unsigned int pixel_y,
+        unsigned int button, unsigned int event, unsigned char flags);
 
 /** @} */
 
