@@ -110,12 +110,18 @@ static void move_cursor(struct tsm_screen *con, unsigned int x, unsigned int y)
 	c->age = con->age_cnt;
 }
 
-void screen_cell_init(struct tsm_screen *con, struct cell *cell)
+void screen_cell_init_generic(struct tsm_screen *con, struct cell *cell, struct tsm_screen_attr *attr)
 {
 	cell->ch = 0;
 	cell->width = 1;
 	cell->age = con->age_cnt;
-	memcpy(&cell->attr, &con->def_attr, sizeof(cell->attr));
+
+	memcpy(&cell->attr, attr, sizeof(cell->attr));
+}
+
+void screen_cell_init(struct tsm_screen *con, struct cell *cell)
+{
+	screen_cell_init_generic(con, cell, &con->def_attr);
 }
 
 static int line_new(struct tsm_screen *con, struct line **out,
@@ -224,10 +230,13 @@ static void link_to_scrollback(struct tsm_screen *con, struct line *line)
 		if (con->sb_pos) {
 			if (con->sb_pos == tmp ||
 			    !(con->flags & TSM_SCREEN_FIXED_POS)) {
-				if (con->sb_pos->next)
+				if (con->sb_pos->next) {
 					con->sb_pos = con->sb_pos->next;
-				else
+					++con->sb_pos_num;
+				} else {
 					con->sb_pos = line;
+					con->sb_pos_num = 0;
+				}
 			}
 		}
 
@@ -247,12 +256,17 @@ static void link_to_scrollback(struct tsm_screen *con, struct line *line)
 	line->sb_id = ++con->sb_last_id;
 	line->next = NULL;
 	line->prev = con->sb_last;
-	if (con->sb_last)
+	if (con->sb_last) {
 		con->sb_last->next = line;
-	else
+	} else {
 		con->sb_first = line;
+	}
 	con->sb_last = line;
 	++con->sb_count;
+
+	if (con->sb_pos == NULL) {
+		con->sb_pos_num = con->sb_count;
+	}
 }
 
 static void screen_scroll_up(struct tsm_screen *con, unsigned int num)
@@ -532,10 +546,12 @@ void tsm_screen_unref(struct tsm_screen *con)
 		line_free(con->main_lines[i]);
 		line_free(con->alt_lines[i]);
 	}
+
 	free(con->main_lines);
 	free(con->alt_lines);
 	free(con->tab_ruler);
 	tsm_symbol_table_unref(con->sym_table);
+	tsm_screen_clear_sb(con);
 	free(con);
 }
 
@@ -658,7 +674,6 @@ int tsm_screen_resize(struct tsm_screen *con, unsigned int x,
 			ret = line_resize(con, con->main_lines[i], x);
 			if (ret)
 				return ret;
-
 			ret = line_resize(con, con->alt_lines[i], x);
 			if (ret)
 				return ret;
@@ -678,7 +693,8 @@ int tsm_screen_resize(struct tsm_screen *con, unsigned int x,
 			i = start;
 
 		for ( ; i < con->main_lines[j]->size; ++i)
-			screen_cell_init(con, &con->main_lines[j]->cells[i]);
+			screen_cell_init_generic(con, &con->main_lines[j]->cells[i],
+				&con->def_attr_main);
 
 		/* alt-lines never go into SB, only clear visible cells */
 		i = 0;
@@ -824,6 +840,7 @@ void tsm_screen_clear_sb(struct tsm_screen *con)
 	con->sb_last = NULL;
 	con->sb_count = 0;
 	con->sb_pos = NULL;
+	con->sb_pos_num = 0;
 
 	if (con->sel_active) {
 		if (con->sel_start.line) {
@@ -853,10 +870,12 @@ void tsm_screen_sb_up(struct tsm_screen *con, unsigned int num)
 				return;
 
 			con->sb_pos = con->sb_pos->prev;
+			--con->sb_pos_num;
 		} else if (!con->sb_last) {
 			return;
 		} else {
 			con->sb_pos = con->sb_last;
+			con->sb_pos_num = con->sb_count - 1;
 		}
 	}
 }
@@ -872,8 +891,10 @@ void tsm_screen_sb_down(struct tsm_screen *con, unsigned int num)
 	con->age = con->age_cnt;
 
 	while (num--) {
-		if (con->sb_pos)
+		if (con->sb_pos) {
 			con->sb_pos = con->sb_pos->next;
+			++con->sb_pos_num;
+		}
 		else
 			return;
 	}
@@ -910,6 +931,25 @@ void tsm_screen_sb_reset(struct tsm_screen *con)
 	con->age = con->age_cnt;
 
 	con->sb_pos = NULL;
+	con->sb_pos_num = 0;
+}
+
+unsigned int tsm_screen_sb_get_line_count(struct tsm_screen *con)
+{
+	if (!con) {
+		return 0;
+	}
+
+	return con->sb_count;
+}
+
+unsigned int tsm_screen_sb_get_line_pos(struct tsm_screen *con)
+{
+	if (!con) {
+		return 0;
+	}
+
+	return con->sb_pos_num;
 }
 
 SHL_EXPORT
@@ -918,7 +958,6 @@ void tsm_screen_set_def_attr(struct tsm_screen *con,
 {
 	if (!con || !attr)
 		return;
-
 	memcpy(&con->def_attr, attr, sizeof(*attr));
 }
 
@@ -963,6 +1002,9 @@ void tsm_screen_set_flags(struct tsm_screen *con, unsigned int flags)
 	if (!(old & TSM_SCREEN_ALTERNATE) && (flags & TSM_SCREEN_ALTERNATE)) {
 		con->age = con->age_cnt;
 		con->lines = con->alt_lines;
+
+		/* save attributes of main screen when we switch to alt screen */
+		memcpy(&con->def_attr_main, &con->def_attr, sizeof(con->def_attr));
 	}
 
 	if (!(old & TSM_SCREEN_HIDE_CURSOR) &&
