@@ -155,6 +155,7 @@ struct tsm_vte {
 	tsm_vte_write_cb write_cb;
 	void *data;
 	char *palette_name;
+	bool backspace_sends_delete;
 
 	struct tsm_utf8_mach *mach;
 	unsigned long parse_cnt;
@@ -169,6 +170,13 @@ struct tsm_vte {
 	unsigned int osc_len;
 	char osc_arg[OSC_MAX_LEN];
 
+	tsm_vte_mouse_cb mouse_cb;
+	void *mouse_data;
+	unsigned int mouse_mode;
+	unsigned int mouse_event;
+	unsigned int mouse_last_col;
+	unsigned int mouse_last_row;
+
 	uint8_t (*custom_palette_storage)[3];
 	uint8_t (*palette)[3];
 	struct tsm_screen_attr def_attr;
@@ -189,7 +197,7 @@ struct tsm_vte {
 	unsigned int alt_cursor_y;
 };
 
-static uint8_t color_palette[TSM_COLOR_NUM][3] = {
+static uint8_t color_palette_legacy[TSM_COLOR_NUM][3] = {
 	[TSM_COLOR_BLACK]         = {   0,   0,   0 }, /* black */
 	[TSM_COLOR_RED]           = { 205,   0,   0 }, /* red */
 	[TSM_COLOR_GREEN]         = {   0, 205,   0 }, /* green */
@@ -211,6 +219,29 @@ static uint8_t color_palette[TSM_COLOR_NUM][3] = {
 	[TSM_COLOR_BACKGROUND]    = {   0,   0,   0 }, /* black */
 };
 
+// vga palette from https://en.wikipedia.org/wiki/ANSI_escape_code
+static uint8_t color_palette_vga[TSM_COLOR_NUM][3] = {
+	[TSM_COLOR_BLACK]         = {   0,   0,   0 }, /* black */
+	[TSM_COLOR_RED]           = { 170,   0,   0 }, /* red */
+	[TSM_COLOR_GREEN]         = {   0, 170,   0 }, /* green */
+	[TSM_COLOR_YELLOW]        = { 170,  85,   0 }, /* yellow */
+	[TSM_COLOR_BLUE]          = {   0,   0, 170 }, /* blue */
+	[TSM_COLOR_MAGENTA]       = { 170,   0, 170 }, /* magenta */
+	[TSM_COLOR_CYAN]          = {   0, 170, 170 }, /* cyan */
+	[TSM_COLOR_LIGHT_GREY]    = { 170, 170, 170 }, /* light grey */
+	[TSM_COLOR_DARK_GREY]     = {  85,  85,  85 }, /* dark grey */
+	[TSM_COLOR_LIGHT_RED]     = { 255,  85,  85 }, /* light red */
+	[TSM_COLOR_LIGHT_GREEN]   = {  85, 255,  85 }, /* light green */
+	[TSM_COLOR_LIGHT_YELLOW]  = { 255, 255,  85 }, /* light yellow */
+	[TSM_COLOR_LIGHT_BLUE]    = {  85,  85, 255 }, /* light blue */
+	[TSM_COLOR_LIGHT_MAGENTA] = { 255,  85, 255 }, /* light magenta */
+	[TSM_COLOR_LIGHT_CYAN]    = {  85, 255, 255 }, /* light cyan */
+	[TSM_COLOR_WHITE]         = { 255, 255, 255 }, /* white */
+
+	[TSM_COLOR_FOREGROUND]    = { 170, 170, 170 }, /* light grey */
+	[TSM_COLOR_BACKGROUND]    = {   0,   0,   0 }, /* black */
+};
+
 static uint8_t color_palette_solarized[TSM_COLOR_NUM][3] = {
 	[TSM_COLOR_BLACK]         = {   7,  54,  66 }, /* black */
 	[TSM_COLOR_RED]           = { 220,  50,  47 }, /* red */
@@ -346,7 +377,7 @@ static uint8_t color_palette_base16_light[TSM_COLOR_NUM][3] = {
 static uint8_t (*get_palette(struct tsm_vte *vte))[3]
 {
 	if (!vte->palette_name)
-		return color_palette;
+		return color_palette_legacy;
 
 	if (!strcmp(vte->palette_name, "custom") && vte->custom_palette_storage)
 		return vte->custom_palette_storage;
@@ -362,8 +393,12 @@ static uint8_t (*get_palette(struct tsm_vte *vte))[3]
 		return color_palette_base16_dark;
 	if (!strcmp(vte->palette_name, "base16-light"))
 		return color_palette_base16_light;
+	if (!strcmp(vte->palette_name, "vga"))
+		return color_palette_vga;
+	if (!strcmp(vte->palette_name, "legacy"))
+		return color_palette_legacy;
 
-	return color_palette;
+	return color_palette_legacy;
 }
 
 /* Several effects may occur when non-RGB colors are used. For instance, if bold
@@ -442,8 +477,11 @@ int tsm_vte_new(struct tsm_vte **out, struct tsm_screen *con,
 	vte->con = con;
 	vte->write_cb = write_cb;
 	vte->data = data;
+	vte->backspace_sends_delete = false;
 	vte->osc_cb = NULL;
 	vte->osc_data = NULL;
+	vte->mouse_cb = NULL;
+	vte->mouse_data = NULL;
 	vte->custom_palette_storage = NULL;
 	vte->palette = get_palette(vte);
 	vte->def_attr.fccode = TSM_COLOR_FOREGROUND;
@@ -503,6 +541,16 @@ void tsm_vte_set_osc_cb(struct tsm_vte *vte, tsm_vte_osc_cb osc_cb, void *osc_da
 	vte->osc_data = osc_data;
 }
 
+SHL_EXPORT
+void tsm_vte_set_mouse_cb(struct tsm_vte *vte, tsm_vte_mouse_cb mouse_cb, void *mouse_data)
+{
+	if (!vte)
+		return;
+
+	vte->mouse_cb = mouse_cb;
+	vte->mouse_data = mouse_data;
+}
+
 static int vte_update_palette(struct tsm_vte *vte)
 {
 	vte->palette = get_palette(vte);
@@ -541,7 +589,7 @@ int tsm_vte_set_palette(struct tsm_vte *vte, const char *palette_name)
 SHL_EXPORT
 int tsm_vte_set_custom_palette(struct tsm_vte *vte, uint8_t (*palette)[3])
 {
-	const size_t palette_byte_size = sizeof(color_palette);
+	const size_t palette_byte_size = sizeof(color_palette_legacy);
 	uint8_t (*tmp)[3] = NULL;
 
 	if (!vte)
@@ -569,12 +617,42 @@ void tsm_vte_get_def_attr(struct tsm_vte *vte, struct tsm_screen_attr *out)
 	memcpy(out, &vte->def_attr, sizeof(*out));
 }
 
+SHL_EXPORT
+unsigned int tsm_vte_get_flags(struct tsm_vte *vte)
+{
+	if (!vte) {
+		return 0;
+	}
+
+	return vte->flags;
+}
+
+SHL_EXPORT
+unsigned int tsm_vte_get_mouse_mode(struct tsm_vte *vte)
+{
+	if (!vte) {
+		return 0;
+	}
+
+	return vte->mouse_mode;
+}
+
+SHL_EXPORT
+unsigned int tsm_vte_get_mouse_event(struct tsm_vte *vte)
+{
+	if (!vte) {
+		return 0;
+	}
+
+	return vte->mouse_event;
+}
+
 /*
  * Write raw byte-stream to pty.
  * When writing data to the client we must make sure that we send the correct
  * encoding. For backwards-compatibility reasons we should always send 7bit
- * characters exclusively. However, when FLAG_7BIT_MODE is not set, then we can
- * also send raw 8bit characters. For instance, in FLAG_8BIT_MODE we can use the
+ * characters exclusively. However, when TSM_VTE_FLAG_7BIT_MODE is not set, then we can
+ * also send raw 8bit characters. For instance, in TSM_VTE_FLAG_8BIT_MODE we can use the
  * GR characters as keyboard input and send them directly or even use the C1
  * escape characters. In unicode mode (default) we can send multi-byte utf-8
  * characters which are also 8bit. When sending these characters, set the \raw
@@ -619,17 +697,17 @@ static void vte_write_debug(struct tsm_vte *vte, const char *u8, size_t len,
 #endif
 
 	/* in local echo mode, directly parse the data again */
-	if (!vte->parse_cnt && !(vte->flags & FLAG_SEND_RECEIVE_MODE)) {
-		if (vte->flags & FLAG_PREPEND_ESCAPE)
+	if (!vte->parse_cnt && !(vte->flags & TSM_VTE_FLAG_SEND_RECEIVE_MODE)) {
+		if (vte->flags & TSM_VTE_FLAG_PREPEND_ESCAPE)
 			tsm_vte_input(vte, "\e", 1);
 		tsm_vte_input(vte, u8, len);
 	}
 
-	if (vte->flags & FLAG_PREPEND_ESCAPE)
+	if (vte->flags & TSM_VTE_FLAG_PREPEND_ESCAPE)
 		vte->write_cb(vte, "\e", 1, vte->data);
 	vte->write_cb(vte, u8, len, vte->data);
 
-	vte->flags &= ~FLAG_PREPEND_ESCAPE;
+	vte->flags &= ~TSM_VTE_FLAG_PREPEND_ESCAPE;
 }
 
 #define vte_write(_vte, _u8, _len) \
@@ -670,8 +748,8 @@ static void save_state(struct tsm_vte *vte)
 	vte->saved_state.cattr = vte->cattr;
 	vte->saved_state.gl = vte->gl;
 	vte->saved_state.gr = vte->gr;
-	vte->saved_state.wrap_mode = vte->flags & FLAG_AUTO_WRAP_MODE;
-	vte->saved_state.origin_mode = vte->flags & FLAG_ORIGIN_MODE;
+	vte->saved_state.wrap_mode = vte->flags & TSM_VTE_FLAG_AUTO_WRAP_MODE;
+	vte->saved_state.origin_mode = vte->flags & TSM_VTE_FLAG_ORIGIN_MODE;
 }
 
 static void restore_state(struct tsm_vte *vte)
@@ -680,24 +758,24 @@ static void restore_state(struct tsm_vte *vte)
 			       vte->saved_state.cursor_y);
 	vte->cattr = vte->saved_state.cattr;
 	to_rgb(vte, &vte->cattr);
-	if (vte->flags & FLAG_BACKGROUND_COLOR_ERASE_MODE)
+	if (vte->flags & TSM_VTE_FLAG_BACKGROUND_COLOR_ERASE_MODE)
 		tsm_screen_set_def_attr(vte->con, &vte->cattr);
 	vte->gl = vte->saved_state.gl;
 	vte->gr = vte->saved_state.gr;
 
 	if (vte->saved_state.wrap_mode) {
-		vte->flags |= FLAG_AUTO_WRAP_MODE;
+		vte->flags |= TSM_VTE_FLAG_AUTO_WRAP_MODE;
 		tsm_screen_set_flags(vte->con, TSM_SCREEN_AUTO_WRAP);
 	} else {
-		vte->flags &= ~FLAG_AUTO_WRAP_MODE;
+		vte->flags &= ~TSM_VTE_FLAG_AUTO_WRAP_MODE;
 		tsm_screen_reset_flags(vte->con, TSM_SCREEN_AUTO_WRAP);
 	}
 
 	if (vte->saved_state.origin_mode) {
-		vte->flags |= FLAG_ORIGIN_MODE;
+		vte->flags |= TSM_VTE_FLAG_ORIGIN_MODE;
 		tsm_screen_set_flags(vte->con, TSM_SCREEN_REL_ORIGIN);
 	} else {
-		vte->flags &= ~FLAG_ORIGIN_MODE;
+		vte->flags &= ~TSM_VTE_FLAG_ORIGIN_MODE;
 		tsm_screen_reset_flags(vte->con, TSM_SCREEN_REL_ORIGIN);
 	}
 }
@@ -715,11 +793,11 @@ void tsm_vte_reset(struct tsm_vte *vte)
 		return;
 
 	vte->flags = 0;
-	vte->flags |= FLAG_TEXT_CURSOR_MODE;
-	vte->flags |= FLAG_AUTO_REPEAT_MODE;
-	vte->flags |= FLAG_SEND_RECEIVE_MODE;
-	vte->flags |= FLAG_AUTO_WRAP_MODE;
-	vte->flags |= FLAG_BACKGROUND_COLOR_ERASE_MODE;
+	vte->flags |= TSM_VTE_FLAG_TEXT_CURSOR_MODE;
+	vte->flags |= TSM_VTE_FLAG_AUTO_REPEAT_MODE;
+	vte->flags |= TSM_VTE_FLAG_SEND_RECEIVE_MODE;
+	vte->flags |= TSM_VTE_FLAG_AUTO_WRAP_MODE;
+	vte->flags |= TSM_VTE_FLAG_BACKGROUND_COLOR_ERASE_MODE;
 	tsm_screen_reset(vte->con);
 	tsm_screen_set_flags(vte->con, TSM_SCREEN_AUTO_WRAP);
 
@@ -734,6 +812,11 @@ void tsm_vte_reset(struct tsm_vte *vte)
 	vte->g2 = &tsm_vte_unicode_lower;
 	vte->g3 = &tsm_vte_unicode_upper;
 
+	vte->mouse_mode = 0;
+	vte->mouse_event = 0;
+	vte->mouse_last_col = 0;
+	vte->mouse_last_row = 0;
+
 	memcpy(&vte->cattr, &vte->def_attr, sizeof(vte->cattr));
 	to_rgb(vte, &vte->cattr);
 	tsm_screen_set_def_attr(vte->con, &vte->def_attr);
@@ -790,7 +873,7 @@ static void do_execute(struct tsm_vte *vte, uint32_t ctrl)
 	case 0x0b: /* VT */
 	case 0x0c: /* FF */
 		/* Line feed or newline (CR/NL mode) */
-		if (vte->flags & FLAG_LINE_FEED_NEW_LINE_MODE)
+		if (vte->flags & TSM_VTE_FLAG_LINE_FEED_NEW_LINE_MODE)
 			tsm_screen_newline(vte->con);
 		else
 			tsm_screen_move_down(vte->con, 1, true);
@@ -1039,7 +1122,7 @@ static void do_esc(struct tsm_vte *vte, uint32_t data)
 		if (vte->csi_flags & CSI_SPACE) {
 			/* S7C1T */
 			/* Disable 8bit C1 mode */
-			vte->flags &= ~FLAG_USE_C1;
+			vte->flags &= ~TSM_VTE_FLAG_USE_C1;
 			return;
 		}
 		break;
@@ -1047,7 +1130,7 @@ static void do_esc(struct tsm_vte *vte, uint32_t data)
 		if (vte->csi_flags & CSI_SPACE) {
 			/* S8C1T */
 			/* Enable 8bit C1 mode */
-			vte->flags |= FLAG_USE_C1;
+			vte->flags |= TSM_VTE_FLAG_USE_C1;
 			return;
 		}
 		break;
@@ -1114,11 +1197,11 @@ static void do_esc(struct tsm_vte *vte, uint32_t data)
 		break;
 	case '=': /* DECKPAM */
 		/* Set application keypad mode */
-		vte->flags |= FLAG_KEYPAD_APPLICATION_MODE;
+		vte->flags |= TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE;
 		break;
 	case '>': /* DECKPNM */
 		/* Set numeric keypad mode */
-		vte->flags &= ~FLAG_KEYPAD_APPLICATION_MODE;
+		vte->flags &= ~TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE;
 		break;
 	case 'c': /* RIS */
 		/* hard reset */
@@ -1360,7 +1443,7 @@ static void csi_attribute(struct tsm_vte *vte)
 	}
 
 	to_rgb(vte, &vte->cattr);
-	if (vte->flags & FLAG_BACKGROUND_COLOR_ERASE_MODE)
+	if (vte->flags & TSM_VTE_FLAG_BACKGROUND_COLOR_ERASE_MODE)
 		tsm_screen_set_def_attr(vte->con, &vte->cattr);
 }
 
@@ -1381,7 +1464,7 @@ static void csi_compat_mode(struct tsm_vte *vte)
 		 * there is no need to explicitly select it.
 		 * However, we enable 7bit mode to avoid
 		 * character-table problems */
-		vte->flags |= FLAG_7BIT_MODE;
+		vte->flags |= TSM_VTE_FLAG_7BIT_MODE;
 		vte->g0 = &tsm_vte_unicode_lower;
 		vte->g1 = &tsm_vte_dec_supplemental_graphics;
 	} else if (vte->csi_argv[0] == 62 ||
@@ -1399,9 +1482,9 @@ static void csi_compat_mode(struct tsm_vte *vte)
 		 * compatibility is requested explicitly. */
 		if (vte->csi_argv[1] == 1 ||
 		    vte->csi_argv[1] == 2)
-			vte->flags |= FLAG_USE_C1;
+			vte->flags |= TSM_VTE_FLAG_USE_C1;
 
-		vte->flags |= FLAG_8BIT_MODE;
+		vte->flags |= TSM_VTE_FLAG_8BIT_MODE;
 		vte->g0 = &tsm_vte_unicode_lower;
 		vte->g1 = &tsm_vte_dec_supplemental_graphics;
 	} else {
@@ -1430,11 +1513,11 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 				continue;
 			case 2: /* KAM */
 				set_reset_flag(vte, set,
-					       FLAG_KEYBOARD_ACTION_MODE);
+					       TSM_VTE_FLAG_KEYBOARD_ACTION_MODE);
 				continue;
 			case 4: /* IRM */
 				set_reset_flag(vte, set,
-					       FLAG_INSERT_REPLACE_MODE);
+					       TSM_VTE_FLAG_INSERT_REPLACE_MODE);
 				if (set)
 					tsm_screen_set_flags(vte->con,
 						TSM_SCREEN_INSERT_MODE);
@@ -1444,11 +1527,11 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 				continue;
 			case 12: /* SRM */
 				set_reset_flag(vte, set,
-					       FLAG_SEND_RECEIVE_MODE);
+					       TSM_VTE_FLAG_SEND_RECEIVE_MODE);
 				continue;
 			case 20: /* LNM */
 				set_reset_flag(vte, set,
-					       FLAG_LINE_FEED_NEW_LINE_MODE);
+					       TSM_VTE_FLAG_LINE_FEED_NEW_LINE_MODE);
 				continue;
 			default:
 				llog_debug(vte, "unknown non-DEC (Re)Set-Mode %d",
@@ -1461,7 +1544,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 		case -1:
 			continue;
 		case 1: /* DECCKM */
-			set_reset_flag(vte, set, FLAG_CURSOR_KEY_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_CURSOR_KEY_MODE);
 			continue;
 		case 2: /* DECANM */
 			/* Select VT52 mode */
@@ -1492,7 +1575,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 			 * scrolling so ignore this here. */
 			continue;
 		case 5: /* DECSCNM */
-			set_reset_flag(vte, set, FLAG_INVERSE_SCREEN_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_INVERSE_SCREEN_MODE);
 			if (set)
 				tsm_screen_set_flags(vte->con,
 						TSM_SCREEN_INVERSE);
@@ -1501,7 +1584,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						TSM_SCREEN_INVERSE);
 			continue;
 		case 6: /* DECOM */
-			set_reset_flag(vte, set, FLAG_ORIGIN_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_ORIGIN_MODE);
 			if (set)
 				tsm_screen_set_flags(vte->con,
 						TSM_SCREEN_REL_ORIGIN);
@@ -1510,7 +1593,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						TSM_SCREEN_REL_ORIGIN);
 			continue;
 		case 7: /* DECAWN */
-			set_reset_flag(vte, set, FLAG_AUTO_WRAP_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_AUTO_WRAP_MODE);
 			if (set)
 				tsm_screen_set_flags(vte->con,
 						TSM_SCREEN_AUTO_WRAP);
@@ -1519,7 +1602,15 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						TSM_SCREEN_AUTO_WRAP);
 			continue;
 		case 8: /* DECARM */
-			set_reset_flag(vte, set, FLAG_AUTO_REPEAT_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_AUTO_REPEAT_MODE);
+			continue;
+		case TSM_VTE_MOUSE_MODE_X10:
+			vte->mouse_mode = set ? vte->csi_argv[i] : 0;
+			vte->mouse_event = TSM_VTE_MOUSE_EVENT_BTN;
+
+			if (vte->mouse_cb) {
+			    vte->mouse_cb(vte, vte->mouse_event, false, vte->mouse_data);
+			}
 			continue;
 		case 12: /* blinking cursor */
 			/* TODO: implement */
@@ -1536,7 +1627,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 			 * this mode. */
 			continue;
 		case 25: /* DECTCEM */
-			set_reset_flag(vte, set, FLAG_TEXT_CURSOR_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_TEXT_CURSOR_MODE);
 			if (set)
 				tsm_screen_reset_flags(vte->con,
 						TSM_SCREEN_HIDE_CURSOR);
@@ -1545,10 +1636,10 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						TSM_SCREEN_HIDE_CURSOR);
 			continue;
 		case 42: /* DECNRCM */
-			set_reset_flag(vte, set, FLAG_NATIONAL_CHARSET_MODE);
+			set_reset_flag(vte, set, TSM_VTE_FLAG_NATIONAL_CHARSET_MODE);
 			continue;
 		case 47: /* Alternate screen buffer */
-			if (vte->flags & FLAG_TITE_INHIBIT_MODE)
+			if (vte->flags & TSM_VTE_FLAG_TITE_INHIBIT_MODE)
 				continue;
 
 			if (set)
@@ -1559,7 +1650,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						       TSM_SCREEN_ALTERNATE);
 			continue;
 		case 1047: /* Alternate screen buffer with post-erase */
-			if (vte->flags & FLAG_TITE_INHIBIT_MODE)
+			if (vte->flags & TSM_VTE_FLAG_TITE_INHIBIT_MODE)
 				continue;
 
 			if (set) {
@@ -1572,7 +1663,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 			}
 			continue;
 		case 1048: /* Set/Reset alternate-screen buffer cursor */
-			if (vte->flags & FLAG_TITE_INHIBIT_MODE)
+			if (vte->flags & TSM_VTE_FLAG_TITE_INHIBIT_MODE)
 				continue;
 
 			if (set) {
@@ -1586,7 +1677,7 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 			}
 			continue;
 		case 1049: /* Alternate screen buffer with pre-erase+cursor */
-			if (vte->flags & FLAG_TITE_INHIBIT_MODE)
+			if (vte->flags & TSM_VTE_FLAG_TITE_INHIBIT_MODE)
 				continue;
 
 			if (set) {
@@ -1604,6 +1695,42 @@ static void csi_mode(struct tsm_vte *vte, bool set)
 						   vte->alt_cursor_y);
 			}
 			continue;
+		case TSM_VTE_MOUSE_EVENT_BTN:
+		case TSM_VTE_MOUSE_EVENT_ANY:
+			if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_X10) {
+			    vte->mouse_event = TSM_VTE_MOUSE_EVENT_BTN;
+			} else {
+			    vte->mouse_event = set ? vte->csi_argv[i] : 0;
+			}
+
+			if (vte->mouse_cb && vte->mouse_mode) {
+			    vte->mouse_cb(vte, vte->mouse_event, vte->mouse_mode == TSM_VTE_MOUSE_MODE_PIXEL, vte->mouse_data);
+			}
+			continue;
+		case TSM_VTE_MOUSE_MODE_SGR:
+			vte->mouse_mode = set ? vte->csi_argv[i] : 0;
+
+			if (!vte->mouse_cb) {
+			    continue;
+			}
+
+			if (!set || vte->mouse_event) {
+			    vte->mouse_cb(vte, vte->mouse_event, false, vte->mouse_data);
+			    continue;
+			}
+			continue;
+		case TSM_VTE_MOUSE_MODE_PIXEL:
+			vte->mouse_mode = set ? vte->csi_argv[i] : 0;
+
+			if (!vte->mouse_cb) {
+			    continue;
+			}
+
+			if (!set || vte->mouse_event) {
+			    vte->mouse_cb(vte, vte->mouse_event, set, vte->mouse_data);
+			    continue;
+			}
+			continue;
 		default:
 			llog_debug(vte, "unknown DEC %set-Mode %d",
 				   set?"S":"Res", vte->csi_argv[i]);
@@ -1691,6 +1818,14 @@ static void do_csi(struct tsm_vte *vte, uint32_t data)
 		x = tsm_screen_get_cursor_x(vte->con);
 		tsm_screen_move_to(vte->con, x, num - 1);
 		break;
+	case 'E': /* CNL */
+		/* Move cursor down the indicated # of rows, to column 1 */
+		num = vte->csi_argv[0];
+		if (num <= 0)
+			num = 1;
+		tsm_screen_move_down(vte->con, num, false);
+		tsm_screen_move_line_home(vte->con);
+		break;
 	case 'e': /* VPR */
 		/* Vertical Line Position Relative */
 		num = vte->csi_argv[0];
@@ -1700,6 +1835,14 @@ static void do_csi(struct tsm_vte *vte, uint32_t data)
 		y = tsm_screen_get_cursor_y(vte->con);
 		tsm_screen_move_to(vte->con, x, y + num);
 		break;
+	case 'F': /* CPL */
+		/* Move cursor up the indicated # of rows, to column 1 */
+		num = vte->csi_argv[0];
+		if (num <= 0)
+			num = 1;
+		tsm_screen_move_up(vte->con, num, false);
+		tsm_screen_move_line_home(vte->con);
+		break;
 	case 'H': /* CUP */
 	case 'f': /* HVP */
 		/* position cursor */
@@ -1774,7 +1917,10 @@ static void do_csi(struct tsm_vte *vte, uint32_t data)
 			/* DECRQM: Request DEC Private Mode */
 			/* If CSI_WHAT is set, then enable,
 			 * otherwise disable */
-			csi_soft_reset(vte);
+
+			/* Ignore */
+
+			/* FIXME: Implement DECRQM */
 		} else {
 			/* DECSCL: Compatibility Level */
 			/* Sometimes CSI_DQUOTE is set here, too */
@@ -2370,12 +2516,12 @@ void tsm_vte_input(struct tsm_vte *vte, const char *u8, size_t len)
 
 	++vte->parse_cnt;
 	for (i = 0; i < len; ++i) {
-		if (vte->flags & FLAG_7BIT_MODE) {
+		if (vte->flags & TSM_VTE_FLAG_7BIT_MODE) {
 			if (u8[i] & 0x80)
 				llog_debug(vte, "receiving 8bit character U+%d from pty while in 7bit mode",
 					   (int)u8[i]);
 			parse_data(vte, u8[i] & 0x7f);
-		} else if (vte->flags & FLAG_8BIT_MODE) {
+		} else if (vte->flags & TSM_VTE_FLAG_8BIT_MODE) {
 			parse_data(vte, u8[i]);
 		} else {
 			state = tsm_utf8_mach_feed(vte->mach, u8[i]);
@@ -2389,6 +2535,12 @@ void tsm_vte_input(struct tsm_vte *vte, const char *u8, size_t len)
 	--vte->parse_cnt;
 }
 
+SHL_EXPORT
+void tsm_vte_set_backspace_sends_delete(struct tsm_vte *vte, bool enable)
+{
+	vte->backspace_sends_delete = enable;
+}
+
 SHL_EXPORT
 bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 			     uint32_t ascii, unsigned int mods,
@@ -2410,7 +2562,7 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 	 * disables this by default, why?) and whether we should implement the
 	 * fallback shifting that xterm does. */
 	if (mods & TSM_ALT_MASK)
-		vte->flags |= FLAG_PREPEND_ESCAPE;
+		vte->flags |= TSM_VTE_FLAG_PREPEND_ESCAPE;
 
 	/* A user might actually use multiple layouts for keyboard input. The
 	 * @keysym variable contains the actual keysym that the user used. But
@@ -2569,7 +2721,10 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 
 	switch (keysym) {
 		case XKB_KEY_BackSpace:
-			vte_write(vte, "\x08", 1);
+			if (vte->backspace_sends_delete)
+				vte_write(vte, "\x7f", 1);
+			else
+				vte_write(vte, "\x08", 1);
 			return true;
 		case XKB_KEY_Tab:
 		case XKB_KEY_KP_Tab:
@@ -2609,13 +2764,13 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 			vte_write(vte, "\x1b", 1);
 			return true;
 		case XKB_KEY_KP_Enter:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE) {
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE) {
 				vte_write(vte, "\eOM", 3);
 				return true;
 			}
 			/* fallthrough */
 		case XKB_KEY_Return:
-			if (vte->flags & FLAG_LINE_FEED_NEW_LINE_MODE)
+			if (vte->flags & TSM_VTE_FLAG_LINE_FEED_NEW_LINE_MODE)
 				vte_write(vte, "\x0d\x0a", 2);
 			else
 				vte_write(vte, "\x0d", 1);
@@ -2644,7 +2799,9 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 		case XKB_KEY_KP_Up:
 			if (mods & TSM_CONTROL_MASK) {
 				vte_write(vte, "\e[1;5A", 6);
-			} else if (vte->flags & FLAG_CURSOR_KEY_MODE) {
+			} else if (mods & TSM_SHIFT_MASK) {
+				vte_write(vte, "\e[1;2A", 6);
+			} else if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE) {
 				vte_write(vte, "\eOA", 3);
 			} else {
 				vte_write(vte, "\e[A", 3);
@@ -2654,7 +2811,9 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 		case XKB_KEY_KP_Down:
 			if (mods & TSM_CONTROL_MASK) {
 				vte_write(vte, "\e[1;5B", 6);
-			} else if (vte->flags & FLAG_CURSOR_KEY_MODE) {
+			} else if (mods & TSM_SHIFT_MASK) {
+				vte_write(vte, "\e[1;2B", 6);
+			} else if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE) {
 				vte_write(vte, "\eOB", 3);
 			} else {
 				vte_write(vte, "\e[B", 3);
@@ -2664,7 +2823,9 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 		case XKB_KEY_KP_Right:
 			if (mods & TSM_CONTROL_MASK) {
 				vte_write(vte, "\e[1;5C", 6);
-			} else if (vte->flags & FLAG_CURSOR_KEY_MODE) {
+			} else if (mods & TSM_SHIFT_MASK) {
+				vte_write(vte, "\e[1;2C", 6);
+			} else if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE) {
 				vte_write(vte, "\eOC", 3);
 			} else {
 				vte_write(vte, "\e[C", 3);
@@ -2674,7 +2835,9 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 		case XKB_KEY_KP_Left:
 			if (mods & TSM_CONTROL_MASK) {
 				vte_write(vte, "\e[1;5D", 6);
-			} else if (vte->flags & FLAG_CURSOR_KEY_MODE) {
+			} else if (mods & TSM_SHIFT_MASK) {
+				vte_write(vte, "\e[1;2D", 6);
+			} else if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE) {
 				vte_write(vte, "\eOD", 3);
 			} else {
 				vte_write(vte, "\e[D", 3);
@@ -2682,113 +2845,113 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 			return true;
 		case XKB_KEY_KP_Insert:
 		case XKB_KEY_KP_0:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOp", 3);
 			else
 				vte_write(vte, "0", 1);
 			return true;
 		case XKB_KEY_KP_1:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOq", 3);
 			else
 				vte_write(vte, "1", 1);
 			return true;
 		case XKB_KEY_KP_2:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOr", 3);
 			else
 				vte_write(vte, "2", 1);
 			return true;
 		case XKB_KEY_KP_3:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOs", 3);
 			else
 				vte_write(vte, "3", 1);
 			return true;
 		case XKB_KEY_KP_4:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOt", 3);
 			else
 				vte_write(vte, "4", 1);
 			return true;
 		case XKB_KEY_KP_5:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOu", 3);
 			else
 				vte_write(vte, "5", 1);
 			return true;
 		case XKB_KEY_KP_6:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOv", 3);
 			else
 				vte_write(vte, "6", 1);
 			return true;
 		case XKB_KEY_KP_7:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOw", 3);
 			else
 				vte_write(vte, "7", 1);
 			return true;
 		case XKB_KEY_KP_8:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOx", 3);
 			else
 				vte_write(vte, "8", 1);
 			return true;
 		case XKB_KEY_KP_9:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOy", 3);
 			else
 				vte_write(vte, "9", 1);
 			return true;
 		case XKB_KEY_KP_Subtract:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOm", 3);
 			else
 				vte_write(vte, "-", 1);
 			return true;
 		case XKB_KEY_KP_Separator:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOl", 3);
 			else
 				vte_write(vte, ",", 1);
 			return true;
 		case XKB_KEY_KP_Delete:
 		case XKB_KEY_KP_Decimal:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOn", 3);
 			else
 				vte_write(vte, ".", 1);
 			return true;
 		case XKB_KEY_KP_Equal:
 		case XKB_KEY_KP_Divide:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOj", 3);
 			else
 				vte_write(vte, "/", 1);
 			return true;
 		case XKB_KEY_KP_Multiply:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOo", 3);
 			else
 				vte_write(vte, "*", 1);
 			return true;
 		case XKB_KEY_KP_Add:
-			if (vte->flags & FLAG_KEYPAD_APPLICATION_MODE)
+			if (vte->flags & TSM_VTE_FLAG_KEYPAD_APPLICATION_MODE)
 				vte_write(vte, "\eOk", 3);
 			else
 				vte_write(vte, "+", 1);
 			return true;
 		case XKB_KEY_Home:
 		case XKB_KEY_KP_Home:
-			if (vte->flags & FLAG_CURSOR_KEY_MODE)
+			if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE)
 				vte_write(vte, "\eOH", 3);
 			else
 				vte_write(vte, "\e[H", 3);
 			return true;
 		case XKB_KEY_End:
 		case XKB_KEY_KP_End:
-			if (vte->flags & FLAG_CURSOR_KEY_MODE)
+			if (vte->flags & TSM_VTE_FLAG_CURSOR_KEY_MODE)
 				vte_write(vte, "\eOF", 3);
 			else
 				vte_write(vte, "\e[F", 3);
@@ -2937,7 +3100,7 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 	}
 
 	if (unicode != TSM_VTE_INVALID) {
-		if (vte->flags & FLAG_7BIT_MODE) {
+		if (vte->flags & TSM_VTE_FLAG_7BIT_MODE) {
 			val = unicode;
 			if (unicode & 0x80) {
 				llog_debug(vte, "invalid keyboard input in 7bit mode U+%x; mapping to '?'",
@@ -2945,7 +3108,7 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 				val = '?';
 			}
 			vte_write(vte, &val, 1);
-		} else if (vte->flags & FLAG_8BIT_MODE) {
+		} else if (vte->flags & TSM_VTE_FLAG_8BIT_MODE) {
 			val = unicode;
 			if (unicode > 0xff) {
 				llog_debug(vte, "invalid keyboard input in 8bit mode U+%x; mapping to '?'",
@@ -2960,6 +3123,91 @@ bool tsm_vte_handle_keyboard(struct tsm_vte *vte, uint32_t keysym,
 		return true;
 	}
 
-	vte->flags &= ~FLAG_PREPEND_ESCAPE;
+	vte->flags &= ~TSM_VTE_FLAG_PREPEND_ESCAPE;
+	return false;
+}
+
+bool tsm_vte_handle_mouse(struct tsm_vte *vte, unsigned int cell_x,
+		unsigned int cell_y, unsigned int pixel_x, unsigned int pixel_y,
+		unsigned int button, unsigned int event, unsigned char modifiers)
+{
+	char buffer[24];
+	unsigned char reply_flags = 0;
+	bool pressed = event & TSM_MOUSE_EVENT_PRESSED;
+
+	/* drop move event if we don't wait for move events */
+	if ((vte->mouse_mode == TSM_VTE_MOUSE_MODE_X10 || vte->mouse_event != TSM_VTE_MOUSE_EVENT_ANY) && (event & TSM_MOUSE_EVENT_MOVED)) {
+		return false;
+	}
+
+	if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_SGR || vte->mouse_mode == TSM_VTE_MOUSE_MODE_PIXEL) {
+		/* internally we use zero indexing but the xterm spec requires the
+		 * top left cell to have the coordinates 1,1 */
+		cell_x++;
+		cell_y++;
+
+		if (button == TSM_MOUSE_BUTTON_WHEEL_UP) {
+			button = 64;
+		} else if (button == TSM_MOUSE_BUTTON_WHEEL_DOWN) {
+			button = 65;
+		}
+
+		reply_flags = button | modifiers;
+	}
+
+	if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_X10) {
+		/* + 0x20 to start in the range of visible characters
+		 * and + 1 to start at 1,1 */
+		cell_x += 0x21;
+		cell_y += 0x21;
+
+		if (cell_x > 0xff) {
+			cell_x = 0xff;
+		}
+
+		if (cell_y > 0xff) {
+			cell_y = 0xff;
+		}
+
+		if (event & TSM_MOUSE_EVENT_RELEASED) {
+			/* translates to released but the information which key is
+			 * released gets lost by design of this encoding scheme */
+			button = 3;
+		}
+
+		reply_flags = (button | modifiers) + 0x20;
+		snprintf((char*) &buffer, sizeof(buffer), "\e[M%c%c%c", reply_flags, cell_x, cell_y);
+
+		vte_write(vte, buffer, strlen(buffer));
+		return true;
+	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_SGR && vte->mouse_event) {
+		if (event & TSM_MOUSE_EVENT_MOVED) {
+			if (cell_x == vte->mouse_last_col && cell_y == vte->mouse_last_row) {
+				return false;
+			}
+
+			reply_flags = 35;
+			pressed = true;
+
+			vte->mouse_last_col = cell_x;
+			vte->mouse_last_row = cell_y;
+		}
+
+		snprintf((char*) &buffer, sizeof(buffer), "\e[<%d;%d;%d%c", reply_flags, cell_x, cell_y, pressed ? 'M' : 'm');
+
+		vte_write(vte, buffer, strlen(buffer));
+		return true;
+	} else if (vte->mouse_mode == TSM_VTE_MOUSE_MODE_PIXEL && vte->mouse_event) {
+		if (event == TSM_MOUSE_EVENT_MOVED) {
+			reply_flags = 35;
+			pressed = true;
+		}
+
+		snprintf((char*) &buffer, sizeof(buffer), "\e[<%d;%d;%d%c", reply_flags, pixel_x, pixel_y, pressed ? 'M' : 'm');
+
+		vte_write(vte, buffer, strlen(buffer));
+		return true;
+	}
+
 	return false;
 }
